use std::fs::File;
use std::io::BufReader;
use std::sync::Arc;

use actix_web::{App, HttpRequest, HttpResponse, HttpServer, middleware, Responder, web};
use actix_web::web::Data;
use clap::Parser;
use rustls::{Certificate, PrivateKey};
use rustls_pemfile::{certs, Item, read_one};

use base::RelayedPort;
use tcp_relay_server::relay_ws::relay_ws;
use tcp_relay_server::server_config::ServerConfig;
use tcp_relay_server::tls_cert_client_verifier::CustomCertClientVerifier;

pub async fn hello_route() -> &'static str {
    "Hello world!"
}

pub async fn config_route(req: HttpRequest, data: Data<Arc<ServerConfig>>) -> impl Responder {
    if data.has_token_auth() {
        let token = req.headers().get("Authorization")
            .map(|t| t.to_str().unwrap_or_default())
            .unwrap_or_default()
            .strip_prefix("Bearer ")
            .unwrap_or_default();

        if !data.tokens.iter().any(|t| t.eq(token)) {
            return HttpResponse::Unauthorized().json("Missing / invalid token");
        }
    }

    HttpResponse::Ok().json(
        data.ports.iter()
            .enumerate()
            .map(|(id, port)| RelayedPort { id, port: port + data.increment_ports })
            .collect::<Vec<_>>()
    )
}

#[actix_web::main]
async fn main() -> std::io::Result<()> {
    env_logger::init_from_env(env_logger::Env::new().default_filter_or("info"));

    let mut args: ServerConfig = ServerConfig::parse();

    // Check if no port are to be forwarded
    if args.ports.is_empty() {
        log::error!("No port to forward!");
        std::process::exit(2);
    }

    // Read tokens from file, if any
    if let Some(file) = &args.tokens_file {
        std::fs::read_to_string(file)
            .expect("Failed to read tokens file!")
            .split('\n')
            .filter(|l| !l.is_empty())
            .for_each(|t| args.tokens.push(t.to_string()));
    }

    if !args.has_auth() {
        log::error!("No authentication method specified!");
        std::process::exit(3);
    }

    if args.has_tls_client_auth() && !args.has_tls_config() {
        log::error!("Cannot provide client auth without TLS configuration!");
        panic!();
    }

    let args = Arc::new(args);

    // Load TLS configuration, if any
    let tls_config = if let (Some(cert), Some(key)) = (&args.tls_cert, &args.tls_key) {

        // Load TLS certificate & private key
        let cert_file = &mut BufReader::new(File::open(cert).unwrap());
        let key_file = &mut BufReader::new(File::open(key).unwrap());

        // Get certificates chain
        let cert_chain = certs(cert_file).unwrap()
            .into_iter()
            .map(Certificate)
            .collect();

        // Get private key
        let key = match read_one(key_file).expect("Failed to read private key!") {
            None => {
                log::error!("Failed to extract private key!");
                panic!();
            }
            Some(Item::PKCS8Key(key)) => key,
            Some(Item::RSAKey(key)) => key,
            _ => {
                log::error!("Unsupported private key type!");
                panic!();
            }
        };

        let config = rustls::ServerConfig::builder()
            .with_safe_defaults();

        let config = match args.has_tls_client_auth() {
            true => config.with_client_cert_verifier(
                Arc::new(CustomCertClientVerifier::new(args.clone()))),
            false => config.with_no_client_auth()
        };

        let config = config.with_single_cert(cert_chain, PrivateKey(key))
            .expect("Failed to load TLS certificate!");

        Some(config)
    } else { None };

    log::info!("Starting relay on http://{}", args.listen_address);

    let args_clone = args.clone();
    let server = HttpServer::new(move || {
        App::new()
            .wrap(middleware::Logger::default())
            .app_data(Data::new(args_clone.clone()))
            .route("/", web::get().to(hello_route))
            .route("/config", web::get().to(config_route))
            .route("/ws", web::get().to(relay_ws))
    });

    if let Some(tls_conf) = tls_config {
        server.bind_rustls(&args.listen_address, tls_conf)?
    } else {
        server.bind(&args.listen_address)?
    }.run()
        .await
}